/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2011  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#include "tokener.h"

#include <climits>
#include <cctype>
#include <cstring>

using namespace std;

token token::val(const value& f){
  token tok;
  tok.type=TOK_VALUE;
  tok.f=f;
  return tok;
}

token token::op(const opid& op){
  token tok;
  tok.type=TOK_OP;
  tok.o=op;
  return tok;
}

token token::parl(uchar type){
  token tok;
  tok.type=TOK_PAR_L;
  tok.i=type;
  return tok;
}

token token::parr(uchar type){
  token tok;
  tok.type=TOK_PAR_R;
  tok.i=type;
  return tok;
}

#if DEBUG>2
void printtoken(token& tok){
  switch(tok.type){
    case TOK_NONE: printf("none "); break;
    case TOK_PAR_L:
      switch(tok.i){
        case OP_LIST: printf("{ ");break;
        case OP_ITER1: printf("[ ");break;
        case OP_ITER2: printf("( ");break;
        case OP_FUN: printf("{: ");break;
      }
      break;
    case TOK_PAR_R:
      switch(tok.i){
        case OP_LIST: printf("} ");break;
        case OP_ITER1: printf("] ");break;
        case OP_ITER2: printf(") ");break;
        case OP_FUN: printf(":} ");break;
      }
      break;
    case TOK_OP:
      if(tok.o.code==OP_MINUSSIGN)
        printf("-");
      else
        printf("%s ",tok.o.name);
      break;
    case TOK_VALUE:
      tok.f.print(stdout,"%d ","%g ");
      break;
  }
}
#endif

void tokenize(char* source, vector<token>& tokens){
  uint seek=0;//string position (input is already separated into lines)

  int whitespace=1;//flag if previous is whitespace

  //fprintf(stderr,"sizeof(value)=%d\n",sizeof(value));
  //fprintf(stderr,"sizeof(opid)=%d\n",sizeof(opid));
  //fprintf(stderr,"sizeof(token)=%d\n",sizeof(token));

  for(int c=source[seek];c;c=source[++seek]){
    #if DEBUG>2
    fprintf(stderr,"parser: pos %d, char '%c'\n",seek,c);
    #endif
    if(isspace(c)||(c==',')){
      whitespace=1;
      continue;
    }
    if(c=='-'){//minus
      if(tokens.empty()){//unary minus if at the beginning
	tokens.push_back(token::op(shared::shared_ops[4]));
	goto end_of_loop;
      }
      if(whitespace&&((!isspace(source[seek+1]))&&(source[seek+1]!=',')&&(source[seek+1]!='-'))){//unary minus if preceded by space, but followed by expression
	tokens.push_back(token::op(shared::shared_ops[4]));
	goto end_of_loop;
      }
      token last=tokens.back();
      //after closing brackets and operators that don't need right argument, this is unary minus
      if(((last.type==TOK_OP) && (last.o.asoc!=ASOC_LUNARY) && (last.o.asoc!=ASOC_BLANK))||last.type==TOK_PAR_L){
	tokens.push_back(token::op(shared::shared_ops[4]));
	goto end_of_loop;
      }
      goto check_ops;
    }
    if(c=='+'){//plus
      if(tokens.empty())goto end_of_loop;//skip leading plus
      if(whitespace&&((!isspace(source[seek+1]))&&(source[seek+1]!=',')&&(source[seek+1]!='+')))goto end_of_loop;
      token last=tokens.back();
      if(((last.type==TOK_OP) && (last.o.asoc!=ASOC_LUNARY) && (last.o.asoc!=ASOC_BLANK))||last.type==TOK_PAR_L)goto end_of_loop;
      goto check_ops;
    }
    //delims
    if(c=='{' && source[seek+1]==':'){
      tokens.push_back(token::parl(OP_FUN));
      seek++;
      goto end_of_loop;
    }
    if(c==':' && source[seek+1]=='}'){
      tokens.push_back(token::parr(OP_FUN));
      seek++;
      goto end_of_loop;
    }
    if(c=='{'){
      tokens.push_back(token::parl(OP_LIST));
      goto end_of_loop;
    }
    if(c=='}'){
      tokens.push_back(token::parr(OP_LIST));
      goto end_of_loop;
    }
    if(c=='['){
      tokens.push_back(token::parl(OP_ITER1));
      goto end_of_loop;
    }
    if(c==']'){
      tokens.push_back(token::parr(OP_ITER1));
      goto end_of_loop;
    }
    if(c=='('){
      tokens.push_back(token::parl(OP_ITER2));
      goto end_of_loop;
    }
    if(c==')'){
      tokens.push_back(token::parr(OP_ITER2));
      goto end_of_loop;
    }
    //check for number
    if(isdigit(c)||(c=='.')){//a number encountered
      char* currentstring=source+seek;
      seek++;
      while(isdigit(source[seek])||source[seek]=='.')seek++;
      //remember what was there
      char c_temp=source[seek];
      source[seek]=0;//fake end of string
      tokens.push_back(token::val(value(currentstring)));//parse what was there
#if DEBUG>2
      fprintf(stderr,"parser: value string found: '%s'\n",currentstring);
#endif
      source[seek]=c_temp;//restore original string
      //we end up at the end of the read number
      seek--;//seek++ must be compensated
      goto end_of_loop;
    }
  check_ops:
    {//parse operators
      opid o=findop(source+seek);
      if(o.code!=OP_NONE){
	tokens.push_back(token::op(o));
	seek+=o.len-1;
	goto end_of_loop;
      }
    }
    //nothing matched
    fprintf(stderr,"parser: skipping bogus character '%c'\n",c);
  end_of_loop:
    whitespace=0;
  }
}

